from __future__ import annotations

import bz2
import logging
import os

import requests

from vunnel import utils
from vunnel.utils.oval_parser import Config, parse


class Parser:
    _url_mappings_ = [  # noqa: RUF012
        # Legacy data for RHEL:5 - no longer available from endpoint after 1st July, 2023; however, it is available in the
        # preload archive.
        {
            "base_url": "https://www.redhat.com/security/data/oval",
            "manifest_path": "PULP_MANIFEST",
            "oval_paths": ["com.redhat.rhsa-all.xml.bz2"],
            "skip_download": True,
        },
        {
            "base_url": "https://www.redhat.com/security/data/oval/v2",
            "manifest_path": "PULP_MANIFEST",
            "oval_paths": [
                "RHEL6/rhel-6.oval.xml.bz2",
                "RHEL7/rhel-7.oval.xml.bz2",
                "RHEL8/rhel-8.oval.xml.bz2",
                "RHEL9/rhel-9.oval.xml.bz2",
            ],
        },
    ]

    def __init__(self, workspace, config: Config, logger=None, download_timeout=125):
        self.config = config
        self.download_timeout = download_timeout
        self.workspace = workspace

        if not logger:
            logger = logging.getLogger(self.__class__.__name__)

        self.logger = logger
        self._urls = set()

    @property
    def urls(self) -> list[str]:
        return list(self._urls)

    def _get_workspace_xml_filepath(self, oval_url_path: str) -> str:
        return os.path.join(self.workspace.input_path, oval_url_path.rstrip(".bz2"))

    @utils.retry_with_backoff()
    def _download_oval_file(self, base_url: str, oval_url_path: str):
        xml_file_path = self._get_workspace_xml_filepath(oval_url_path=oval_url_path)
        try:
            oval_url = f"{base_url}/{oval_url_path}"
            self.logger.info(f"begin downloading OVAL file from {oval_url}")
            r = requests.get(oval_url, stream=True, timeout=self.download_timeout)
            if r.status_code == 200:
                # ensure any nested directories get created
                os.makedirs(os.path.dirname(xml_file_path), exist_ok=True)
                with open(xml_file_path, "wb") as extracted:
                    decompressor = bz2.BZ2Decompressor()
                    for chunk in r.iter_content(chunk_size=1024):
                        uncchunk = decompressor.decompress(chunk)
                        extracted.write(uncchunk)

                self.logger.info(f"finish downloading OVAL file from {oval_url}")
            else:  # noqa: RET505
                raise Exception(f"GET {oval_url} failed with HTTP error {r.status_code}")
        except Exception:
            self.logger.exception("error downloading OVAL file")
            raise Exception("error downloading OVAL file")  # noqa: B904

    @utils.retry_with_backoff()
    def _download(self):
        for m in self._url_mappings_:
            base_url = m["base_url"]
            manifest_path = m["manifest_path"]
            oval_paths = m["oval_paths"]
            skip_download = m.get("skip_download", False)
            self._urls.add(f"{base_url}/{manifest_path}")

            for p in oval_paths:
                self._urls.add(f"{base_url}/{p}")
                if not skip_download:
                    self._download_oval_file(base_url, p)

    def xml_paths(self):
        paths = []
        for m in self._url_mappings_:
            oval_paths = m["oval_paths"]
            skip_download = m.get("skip_download", False)

            for p in oval_paths:
                # normalize and return results
                file_path = self._get_workspace_xml_filepath(p)
                if skip_download and not os.path.exists(file_path):
                    self.logger.warning(f"skip processing OVAL file {p}")
                    continue
                paths.append(file_path)
        return paths

    def parse(self):
        vuln_dict = {}

        for file_path in self.xml_paths():
            self.logger.info(f"begin parsing OVAL file {file_path}")
            partial_results = parse(file_path, self.config, vuln_dict=vuln_dict)
            vuln_dict.update(partial_results)
            self.logger.info(f"finish parsing OVAL file {file_path}")

        return vuln_dict

    def get(self):
        self._download()
        return self.parse()



from __future__ import annotations

import abc
import datetime
import enum
import logging
import os
import tempfile
import time
from dataclasses import dataclass, field
from typing import Any, Optional
from urllib.parse import urlparse

from vunnel.utils import archive, hasher, http

from . import distribution, result, workspace
from . import schema as schema_def
from .result import ResultStatePolicy


class OnErrorAction(str, enum.Enum):
    FAIL = "fail"
    SKIP = "skip"
    RETRY = "retry"

    def __repr__(self) -> str:
        return self.value


class InputStatePolicy(str, enum.Enum):
    KEEP = "keep"
    DELETE = "delete"

    def __repr__(self) -> str:
        return self.value


@dataclass
class OnErrorConfig:
    # the action to take when an error occurs
    action: OnErrorAction = OnErrorAction.FAIL
    # the number of times to retry an action that fails
    retry_count: int = 3
    # the number of seconds to wait between retries
    retry_delay: int = 5
    # what to do with the input directory when an error occurs
    input: InputStatePolicy = InputStatePolicy.KEEP  # noqa: A003
    # what to do with the result directory when an error occurs
    results: ResultStatePolicy = ResultStatePolicy.KEEP

    def __post_init__(self) -> None:
        if not isinstance(self.action, OnErrorAction):
            self.action = OnErrorAction(self.action)
        if not isinstance(self.input, InputStatePolicy):
            self.input = InputStatePolicy(self.input)
        if not isinstance(self.results, ResultStatePolicy):
            self.results = ResultStatePolicy(self.results)


@dataclass
class RuntimeConfig:
    # what to do when an error occurs while the provider is running
    on_error: OnErrorConfig = field(default_factory=OnErrorConfig)
    # what to do with existing data in the input directory before running
    existing_input: InputStatePolicy = InputStatePolicy.KEEP
    # what to do with existing data in the result directory before running
    existing_results: ResultStatePolicy = ResultStatePolicy.KEEP
    # the format the results should be written in
    result_store: result.StoreStrategy = result.StoreStrategy.FLAT_FILE
    # skip checks for newer archive if true (always download latest)
    skip_newer_archive_check: bool = False

    import_results_host: Optional[str] = None  # noqa: UP007 - breaks mashumaro
    import_results_path: Optional[str] = None  # noqa: UP007 - breaks mashumaro
    import_results_enabled: Optional[bool] = None  # noqa: UP007 - breaks mashumaro

    def __post_init__(self) -> None:
        if not isinstance(self.existing_input, InputStatePolicy):
            self.existing_input = InputStatePolicy(self.existing_input)
        if not isinstance(self.existing_results, ResultStatePolicy):
            self.existing_results = ResultStatePolicy(self.existing_results)
        if not isinstance(self.result_store, result.StoreStrategy):
            self.result_store = result.StoreStrategy(self.result_store)

    @property
    def skip_if_exists(self) -> bool:
        return self.existing_input == InputStatePolicy.KEEP

    def import_url(self, provider_name: str) -> str:
        path = self.import_results_path
        if path is None:
            path = ""
        path = path.format(provider_name=provider_name)
        host = self.import_results_host
        if host is None:
            host = ""

        return f"{host.strip('/')}/{path.strip('/')}"


def disallow_existing_input_policy(cfg: RuntimeConfig) -> None:
    if cfg.existing_input != InputStatePolicy.KEEP:
        raise ValueError(
            f"existing_input policy is '{cfg.existing_input}' but only a value of 'keep' is allowed for this provider",
        )


class Provider(abc.ABC):
    # a breaking change to the semantics of how the provider processes results.
    #
    # NOTE: this value should only be changed in classes that inherit this class. Do not change the value in this class!
    __version__: int = 1

    # a breaking change to the schema of the results that the provider writes out should incur a version bump here.
    #
    # NOTE: this value should only be changed in classes that inherit this class. Do not change the value in this class!
    __distribution_version__: int = 1

    def __init__(self, root: str, runtime_cfg: RuntimeConfig = RuntimeConfig()):  # noqa: B008
        self.logger = logging.getLogger(self.name())
        self.workspace = workspace.Workspace(root, self.name(), logger=self.logger, create=False)
        self.urls: list[str] = []
        if runtime_cfg.import_results_enabled:
            if not runtime_cfg.import_results_host:
                raise RuntimeError("enabling import results requires host")
            if not runtime_cfg.import_results_path:
                raise RuntimeError("enabling import results requires path")

        self.runtime_cfg = runtime_cfg

    @classmethod
    def version(cls) -> int:
        return cls.__version__ + (cls.distribution_version() - 1)

    @classmethod
    def distribution_version(cls) -> int:
        """This version represents when a breaking change is made for interpreting purely the provider results. This
        tends to be an aggregation of all schema versions involved in the provider (i.e. the provider workspace state
        and results shape). This is slightly different from the `version` method which is specific to the provider,
        which encapsulates at least the distribution version + any other behavioral or data differences of the
        provider itself (which is valid during processing, but not strictly interpreting results)."""
        workspace_version = int(schema_def.ProviderStateSchema().major_version)
        return (workspace_version - 1) + cls.__distribution_version__

    @classmethod
    @abc.abstractmethod
    def name(cls) -> str:
        # note: wrapping [abstractmethod -> classmethod -> property] is no longer supported in
        # python 3.11 (deprecated due to undefined behavior) for this reason a simple classmethod
        # is used instead of a property
        raise NotImplementedError("'name()' must be implemented")

    @abc.abstractmethod
    def update(self, last_updated: datetime.datetime | None) -> tuple[list[str], int]:
        """Populates the input directory from external sources, processes the data, places results into the output directory."""
        raise NotImplementedError("'update()' must be implemented")

    def read_state(self) -> workspace.State | None:
        try:
            return workspace.State.read(root=self.workspace.path)
        except FileNotFoundError:
            return None

    def _update(self) -> None:
        start = datetime.datetime.now(tz=datetime.timezone.utc)

        last_updated = None
        current_state = self.read_state()
        if current_state and not current_state.stale:
            last_updated = current_state.timestamp

        stale = False
        if self.runtime_cfg.import_results_enabled:
            urls, count, start = self._fetch_or_use_results_archive()
            stale = True
        else:
            urls, count = self.update(last_updated=last_updated)

        if count > 0 or stale:
            self.workspace.record_state(
                stale=stale,
                version=self.version(),
                distribution_version=self.distribution_version(),
                timestamp=start,
                urls=urls,
                store=self.runtime_cfg.result_store.value,
            )
        else:
            self.logger.debug("skipping recording of workspace state (no new results found)")

    def _fetch_or_use_results_archive(self) -> tuple[list[str], int, datetime.datetime]:
        listing_doc = self._fetch_listing_document()
        latest_entry = listing_doc.latest_entry(schema_version=self.distribution_version())
        if not latest_entry:
            raise RuntimeError("no listing entry found")
        timestamp = None
        if self.runtime_cfg.skip_newer_archive_check or self._has_newer_archive(latest_entry=latest_entry):
            self.logger.info("fetching latest listing")
            self._prep_workspace_from_listing_entry(entry=latest_entry)
        else:
            # Update the timestamp of the state to the latest entry's built time
            self.logger.info("using existing listing and updating timestamp")
            timestamp = datetime.datetime.fromisoformat(latest_entry.built)

        state = self.workspace.state()
        if not timestamp:
            timestamp = state.timestamp
        return state.urls, state.result_count(self.workspace.path), timestamp

    def _fetch_listing_document(self) -> distribution.ListingDocument:
        url = self.runtime_cfg.import_url(provider_name=self.name())
        resp = http.get(url, logger=self.logger)
        resp.raise_for_status()

        return distribution.ListingDocument.from_dict(resp.json())

    def _has_newer_archive(self, latest_entry: distribution.ListingEntry) -> bool:
        if not os.path.exists(self.workspace.metadata_path):
            return True

        state = self.workspace.state()
        if not state:
            return True

        if state.distribution_version != self.distribution_version():
            return True

        if not state.listing:
            return True

        # note: the checksum is the digest of the checksums file within the archive, which is in the form "algo:value"
        return f"{state.listing.algorithm}:{state.listing.digest}" != latest_entry.enclosed_checksum

    def _prep_workspace_from_listing_entry(self, entry: distribution.ListingEntry) -> None:
        with tempfile.TemporaryDirectory() as temp_dir:
            unarchived_path = _fetch_listing_entry_archive(dest=temp_dir, entry=entry, logger=self.logger)

            temp_ws = workspace.Workspace(unarchived_path, self.name(), logger=self.logger, create=False)

            # validate that the workspace is in a good state
            temp_ws.validate_checksums()

            # then switch the existing workspace to the new one...
            # move the contents of the tmp dir to the workspace destination
            self.workspace.replace_results(temp_workspace=temp_ws)

    def run(self) -> None:
        self.logger.debug(f"using {self.workspace.path!r} as workspace")

        current_state = self.read_state()
        if self.runtime_cfg.import_results_enabled:
            if current_state and current_state.distribution_version != self.distribution_version():
                self.logger.warning(
                    f"provider distribution version has changed from {current_state.distribution_version} to {self.distribution_version()}",
                )
                self.logger.warning("clearing workspace to ensure consistency of existing results")
                self.workspace.clear()
        elif current_state and current_state.version != self.version():
            self.logger.warning(f"provider version has changed from {current_state.version} to {self.version()}")
            self.logger.warning("clearing workspace to ensure consistency of existing input and results")
            self.workspace.clear()
        else:
            if self.runtime_cfg.existing_results == ResultStatePolicy.DELETE:
                self.workspace.clear_results()

            if self.runtime_cfg.existing_input == InputStatePolicy.DELETE:
                self.workspace.clear_input()

        self.workspace.create()

        try:
            self._update()
        except Exception as e:
            self._on_error(e)

    def _on_error(self, e: Exception) -> None:
        self.logger.error(f"error during update: {e}")

        # manage state
        self._on_error_handle_state()

        # manage event
        if self.runtime_cfg.on_error.action == OnErrorAction.FAIL:
            raise e

        if self.runtime_cfg.on_error.action == OnErrorAction.SKIP:
            return

        if self.runtime_cfg.on_error.action == OnErrorAction.RETRY:
            attempt = 1
            last_exception: Exception | None = e
            while attempt <= self.runtime_cfg.on_error.retry_count:
                self.logger.info(f"retrying after error: {e}")
                time.sleep(self.runtime_cfg.on_error.retry_delay)
                try:
                    self._update()
                    last_exception = None
                    break
                except Exception as ex:
                    self.logger.error(f"error during update (attempt {attempt}): {e}")
                    self._on_error_handle_state()
                    last_exception = ex
                    attempt += 1

            if last_exception:
                raise last_exception

    def _on_error_handle_state(self) -> None:
        if self.runtime_cfg.on_error.input == InputStatePolicy.DELETE:
            self.workspace.clear_input()
        if self.runtime_cfg.on_error.results == ResultStatePolicy.DELETE:
            self.workspace.clear_results()

    def __repr__(self) -> str:
        extra = []
        prefix = ""
        if getattr(self, "config", None):
            extra.append(f"config={self.config}")  # type: ignore[attr-defined]
        if extra:
            prefix = ", "
        return f"Provider(name={self.name()}, input={self.workspace.input_path}{prefix}{', '.join(extra)})"

    def results_writer(self, **kwargs: Any) -> result.Writer:
        return result.Writer(
            workspace=self.workspace,
            result_state_policy=self.runtime_cfg.existing_results,
            logger=self.logger,
            store_strategy=self.runtime_cfg.result_store,
            **kwargs,
        )


def _fetch_listing_entry_archive(dest: str, entry: distribution.ListingEntry, logger: logging.Logger) -> str:
    archive_path = os.path.join(dest, os.path.basename(urlparse(entry.url, allow_fragments=False).path))

    # download the URL for the archive
    resp = http.get(entry.url, logger=logger, stream=True)
    resp.raise_for_status()
    logger.debug(f"downloading {entry.url} to {archive_path}")
    with open(archive_path, "wb") as fp:
        for chunk in resp.iter_content(chunk_size=None):
            fp.write(chunk)

    logger.debug(f"validating checksum for {archive_path}")
    hashMethod = hasher.Method.parse(entry.distribution_checksum)
    actual_labeled_digest = hashMethod.digest(archive_path)
    if actual_labeled_digest != entry.distribution_checksum:
        raise ValueError(f"archive checksum mismatch: {actual_labeled_digest} != {entry.distribution_checksum}")

    unarchive_path = os.path.join(dest, "unarchived")
    logger.debug(f"unarchiving {archive_path} to {unarchive_path}")
    archive.extract(archive_path, unarchive_path)

    return unarchive_path



from __future__ import annotations

from dataclasses import asdict, dataclass, field
from operator import itemgetter
from typing import Any

severity_order = {
    "Unknown": 0,
    "Negligible": 1,
    "Low": 2,
    "Medium": 3,
    "High": 4,
    "Critical": 5,
}

vulnerability_element = {
    "Vulnerability": {
        "Severity": None,
        "NamespaceName": None,
        "FixedIn": [],
        "Link": None,
        "Description": "",
        "Metadata": {},
        "Name": None,
        "CVSS": [],
    },
}


def order_payload(payload, feed_type):
    if payload and feed_type:
        if (
            feed_type == "vulnerabilities"
            and "Vulnerability" in payload
            and "FixedIn" in payload["Vulnerability"]
            and payload["Vulnerability"]["FixedIn"]
        ):
            payload["Vulnerability"]["FixedIn"].sort(key=(itemgetter("Name", "Version")))
        elif feed_type == "packages":
            for content in payload.values():
                for value in content.values():
                    if isinstance(value, list):
                        value.sort()
        else:
            pass

    return payload


@dataclass
class AdvisorySummary:
    ID: str
    Link: str


@dataclass
class VendorAdvisory:
    NoAdvisory: bool
    AdvisorySummary: list[AdvisorySummary] | None

    def __post_init__(self):
        if self.AdvisorySummary is None:
            self.AdvisorySummary = []


@dataclass
class FixedIn:
    """
    Class representing a fix record for return back to the service from the driver. The semantics of the version are:
    "None" -> Package is vulnerable and no fix available yet
    ! "None" -> Version of package with a fix for a vulnerability. Assume all older versions of the package are vulnerable.

    """

    Name: str
    NamespaceName: str
    VersionFormat: str
    Version: str
    Module: str | None
    VendorAdvisory: VendorAdvisory | None
    VulnerableRange: str | None = None

    def __post_init__(self):
        if self.Module is None:
            self.Module = ""
        if self.VendorAdvisory is None:
            self.VendorAdvisory = VendorAdvisory(NoAdvisory=False, AdvisorySummary=None)


@dataclass
class CVSSBaseMetrics:
    base_score: float
    exploitability_score: float
    impact_score: float
    base_severity: str


@dataclass
class CVSS:
    version: str
    vector_string: str
    base_metrics: CVSSBaseMetrics
    status: str


@dataclass
class Vulnerability:
    """
    Class representing the record to be returned. Uses strange capitalization
    to be backwards compatible in the json output with previous version of feed data.
    """

    Name: str
    """Name is a str that contains the vulnerability ID"""
    NamespaceName: str
    Description: str
    Severity: str
    Link: str
    CVSS: list[CVSS]
    FixedIn: list[FixedIn]
    """
    For a package with no fix, a single FixedIn should be reported with
    "None" as the Version. The name field on a FixedIn is the name of the
    vulnerable package.
    """
    Metadata: dict[str, Any] = field(default_factory=dict)

    def to_payload(self):
        return {"Vulnerability": asdict(self)}


def build_reference_links(vulnerability_id: str) -> list[str] | None:
    if vulnerability_id.startswith("CVE-"):
        return [
            f"https://www.cve.org/CVERecord?id={vulnerability_id}",
            f"https://nvd.nist.gov/vuln/detail/{vulnerability_id}",
        ]

    if vulnerability_id.startswith("GHSA-"):
        return [f"https://github.com/advisories/{vulnerability_id}"]

    return None



from __future__ import annotations

import hashlib
from enum import Enum

import xxhash


class Method(Enum):
    SHA256 = "sha256"
    XXH64 = "xxh64"

    def digest(self, path: str, label: bool = True, size: int = 65536) -> str:
        hasher = self.hasher()
        with open(path, "rb") as f:
            while b := f.read(size):
                hasher.update(b)
        if label:
            return self.value + ":" + hasher.hexdigest()
        return hasher.hexdigest()

    def hasher(self):  # type: ignore[no-untyped-def]
        if self == self.SHA256:
            return hashlib.sha256()
        if self == self.XXH64:
            return xxhash.xxh64()
        raise ValueError(f"unknown digest label: {self.value}")

    @staticmethod
    def parse(value: str) -> Method:
        try:
            return Method(value.lower().replace("-", "").strip().split(":")[0])
        except ValueError:
            raise ValueError(f"unknown digest label: {value}") from None



from __future__ import annotations

import os
from dataclasses import dataclass, field
from typing import TYPE_CHECKING

from vunnel import provider, result, schema

from .parser import Parser

if TYPE_CHECKING:
    import datetime


@dataclass
class Config:
    runtime: provider.RuntimeConfig = field(
        default_factory=lambda: provider.RuntimeConfig(
            result_store=result.StoreStrategy.SQLITE,
            existing_results=result.ResultStatePolicy.DELETE_BEFORE_WRITE,
        ),
    )
    request_timeout: int = 125


class Provider(provider.Provider):

    __schema__ = schema.OSSchema()
    __distribution_version__ = int(__schema__.major_version)

    def __init__(self, root: str, config: Config | None = None):
        if not config:
            config = Config()
        super().__init__(root, runtime_cfg=config.runtime)
        self.config = config
        self.logger.debug(f"config: {self.config}")

        self.parser = Parser(
            workspace=self.workspace,
            download_timeout=self.config.request_timeout,
            logger=self.logger,
        )

        # this provider requires the previous state from former runs
        provider.disallow_existing_input_policy(config.runtime)

    @classmethod
    def name(cls) -> str:
        return "alpine"

    def update(self, last_updated: datetime.datetime | None) -> tuple[list[str], int]:
        with self.results_writer() as writer:
            # TODO: tech debt: on subsequent runs, we should only write new vulns (this currently re-writes all)
            for namespace, vulns in self.parser.get():
                namespace = namespace.lower()
                for vuln_id, record in vulns.items():
                    vuln_id = vuln_id.lower()

                    writer.write(
                        identifier=os.path.join(namespace, vuln_id),
                        schema=self.__schema__,
                        payload=record,
                    )

        return self.parser.urls, len(writer)



from __future__ import annotations

import logging
import os
import re
from collections import namedtuple
from html.parser import HTMLParser

import defusedxml.ElementTree as ET

from vunnel.utils import http, rpm

namespace = "amzn"

AlasSummary = namedtuple("AlasSummary", ["id", "url", "sev", "cves"])
AlasFixedIn = namedtuple("AlasFixedIn", ["pkg", "ver"])

amazon_security_advisories = {
    # '1': 'https://alas.aws.amazon.com/alas.rss',
    "2": "https://alas.aws.amazon.com/AL2/alas.rss",
    "2022": "https://alas.aws.amazon.com/AL2022/alas.rss",
    "2023": "https://alas.aws.amazon.com/AL2023/alas.rss",
}

severity_map = {
    "low": "Low",
    "medium": "Medium",
    "important": "High",
    "critical": "Critical",
}


class Parser:
    _title_pattern_ = re.compile(r"([^\s]+)\s+\(([^\)]+)\):.*")
    _whitespace_pattern_ = re.compile(r"\s")
    _rss_file_name_ = "{}_rss.xml"
    _html_dir_name_ = "{}_html"

    def __init__(  # noqa: PLR0913
        self,
        workspace,
        download_timeout=125,
        security_advisories=None,
        logger=None,
        max_allowed_alas_http_403=25,
    ):
        self.workspace = workspace
        self.version_url_map = security_advisories if security_advisories else amazon_security_advisories
        self.download_timeout = download_timeout
        self.max_allowed_alas_http_403 = max_allowed_alas_http_403
        self.urls = []

        self.alas_403s = []

        if not logger:
            logger = logging.getLogger(self.__class__.__name__)
        self.logger = logger

    def _download_rss(self, rss_url, rss_file):
        try:
            self.logger.info(f"downloading amazon security advisory from {rss_url}")
            self.urls.append(rss_url)
            r = http.get(rss_url, self.logger, timeout=self.download_timeout)
            with open(rss_file, "w", encoding="utf-8") as fp:
                fp.write(r.text)
        except Exception:
            self.logger.exception("error downloading amazon linux vulnerability feeds")
            raise

    def _parse_rss(self, file_path):
        self.logger.debug(f"parsing RSS data from {file_path}")
        alas_summaries = []

        processing = False
        for event, element in ET.iterparse(file_path, events=("start", "end")):
            if event == "start" and element.tag == "item":
                processing = True
            elif processing and event == "end":
                if element.tag == "title":
                    found = re.search(self._title_pattern_, element.text.strip())
                    alas_id = found.group(1)
                    sev = found.group(2)
                elif element.tag == "description":
                    desc_str = element.text.strip()
                    cves = re.sub(self._whitespace_pattern_, "", desc_str).split(",") if desc_str else []
                elif element.tag == "link":
                    url = element.text.strip()
                elif element.tag == "item":
                    alas_summaries.append(AlasSummary(id=alas_id, url=url, sev=sev, cves=cves))
                    processing = False

            # clear the element if its not being processed
            if not processing and event == "end":
                element.clear()

        return sorted(alas_summaries)

    def _alas_response_handler(self, response):
        if response.status_code == 403:
            self.alas_403s.append(response.url)
            self.logger.warning(f"403 Forbidden: {response.url}")
        else:
            response.raise_for_status()

    def _get_alas_html(self, alas_url, alas_file, skip_if_exists=True) -> str | None:
        # attempt to download the alas html content
        # if there is a 403, we will skip the download and track the url in self.alas_403s
        # otherwise we will raise the exception
        if skip_if_exists and os.path.exists(alas_file):  # read alas from disk if its available
            self.logger.debug(f"loading existing ALAS from {alas_file}")
            with open(alas_file, encoding="utf-8") as fp:
                content = fp.read()
            return content  # noqa: RET504
        try:
            r = http.get(alas_url, self.logger, timeout=self.download_timeout, status_handler=self._alas_response_handler)
            if r.status_code == 403:
                if len(self.alas_403s) > self.max_allowed_alas_http_403:
                    raise ValueError(
                        f"exceeded maximum allowed 403 responses ({self.max_allowed_alas_http_403}) from ALAS requests",
                    )

                return None
            content = r.text
            with open(alas_file, "w", encoding="utf-8") as fp:
                fp.write(content)
            return content
        except Exception:
            self.logger.exception(f"error downloading data from {alas_url}")
            raise

    @staticmethod
    def get_package_name_version(pkg):
        if not pkg or not isinstance(pkg, str):
            raise ValueError(f"Invalid package name: {pkg}")

        if not pkg.endswith(".rpm"):
            pkg = pkg + ".rpm"

        name, version, release, epoch, arch = rpm.split_rpm_filename(pkg)

        if release:
            return AlasFixedIn(pkg=name, ver=(version + "-" + release))
        return AlasFixedIn(pkg=name, ver=version)

    def get(self, skip_if_exists=False):
        try:
            yield from self._get(skip_if_exists)
        finally:
            if self.alas_403s:
                self.logger.warning(f"failed to fetch {len(self.alas_403s)} ALAS entries due to HTTP 403 response code")
                for url in self.alas_403s:
                    self.logger.warning(f" - {url}")

    def _get(self, skip_if_exists):
        for version, url in self.version_url_map.items():
            rss_file = os.path.join(self.workspace.input_path, self._rss_file_name_.format(version))
            html_dir = os.path.join(self.workspace.input_path, self._html_dir_name_.format(version))

            self._download_rss(url, rss_file)

            # parse rss for alas summaries
            alas_summaries = self._parse_rss(rss_file)

            # setup directory for alas htmls
            if not os.path.exists(html_dir):
                self.logger.debug("initializing workspace for ALAS files")
                os.makedirs(html_dir)

            # iterate through list of alas summaries
            for alas in alas_summaries:
                # download alas html content
                alas_file = os.path.join(html_dir, alas.id)
                html_content = self._get_alas_html(alas.url, alas_file, skip_if_exists=skip_if_exists)

                if html_content is None:
                    self.logger.warning(f"skipping {alas.id}")
                    continue

                # parse alas html for fixes
                parser = PackagesHTMLParser()
                parser.feed(html_content)

                # split the package name and version of the fixed in packages and construct a set
                fixed_in = {self.get_package_name_version(pkg_name) for pkg_name in parser.fixes}

                # concat the descriptions paragraph
                description = "".join(parser.issue_overview_text)

                # construct a vulnerability object and yield it
                yield map_to_vulnerability(version, alas, fixed_in, description)


class JsonifierMixin:
    def json(self):
        jsonified = {}
        for k, v in sorted(vars(self).items()):
            if k[0] != "_":
                if isinstance(v, (list, set)):
                    jsonified[k] = [x.json() if hasattr(x, "json") and callable(x.json) else x for x in v]
                elif isinstance(v, dict):
                    jsonified[k] = {x: y.json() if hasattr(y, "json") and callable(y.json) else y for x, y in sorted(v.items())}
                elif hasattr(v, "json"):
                    jsonified[k] = v.json()
                else:
                    jsonified[k] = v
        return jsonified


class Vulnerability(JsonifierMixin):
    """
    Class representing the record to be returned. Uses strange capitalization
    to be backwards compatible in the json output with previous version of feed data.
    """

    def __init__(self):
        self.Name = None
        self.NamespaceName = None
        self.Description = ""
        self.Severity = None
        self.Metadata = None
        self.Link = None
        self.FixedIn = []


class FixedIn(JsonifierMixin):
    """
    Class representing a fix record for return back to the service from the driver. The semantics of the version are:
    "None" -> Package is vulnerable and no fix available yet
    ! "None" -> Version of package with a fix for a vulnerability. Assume all older versions of the package are vulnerable.

    """

    def __init__(self):
        self.Name = None
        self.NamespaceName = None
        self.VersionFormat = None
        self.Version = None


class PackagesHTMLParser(HTMLParser):
    _new_packages_tuple_ = ("id", "new_packages")
    _arch_list_ = ["x86_64:", "noarch:", "src:"]  # noqa: RUF012

    def __init__(self):
        self.fixes = []
        self.issue_overview_text = []
        self.issue_overview_tag = None
        self.issue_overview_hit = False
        self.fix_tag = None
        self.fix_hit = False
        self.arch_hit = False
        super().__init__()

    def handle_starttag(self, tag, attrs):
        if attrs and self._new_packages_tuple_ in attrs:
            # print('Encountered element with ID new_packages, start tag: {}'.format(tag))
            self.fix_hit = True
            self.fix_tag = tag
        if tag == "div" and ("id", "issue_overview") in attrs:
            self.issue_overview_hit = True
            self.issue_overview_tag = tag
        # else:
        #     print('Ignoring start tag: {}'.format(tag))

    def handle_endtag(self, tag):
        if self.fix_hit and self.fix_tag == tag:
            # print('Encountered end tag for element with ID new_packages')
            self.fix_hit = False
            self.arch_hit = False
        if self.issue_overview_hit and self.issue_overview_tag == tag:
            self.issue_overview_hit = False
        # else:
        #     print('Ignoring end tag: {}'.format(tag))

    def handle_data(self, data):
        data = data.strip()

        if self.fix_hit and data:
            if data in self._arch_list_:  # check if its a relevant arch
                # print('Found relevant arch: "{}"'.format(data))
                self.arch_hit = True
            elif data.endswith(":"):  # Using i686: or src: as an indicator for end of processing
                # if data != 'New Packages:':
                #     logger.debug('Ignoring irrelevant arch or data: {}'.format(data))
                self.arch_hit = False
            elif self.arch_hit:  # check if its a relevant package
                # print('Found relevant package: {}'.format(data))
                self.fixes.append(data)

        if self.issue_overview_hit and data and not data.__contains__("Issue Overview:"):
            self.issue_overview_text.append(data)
        # else:
        #     print('Ignoring data: {}'.format(data.strip()))


def map_to_vulnerability(version, alas, fixed_in, description):
    if not alas:
        raise ValueError("Invalid reference to AlasSummary")

    v = Vulnerability()
    v.Name = alas.id
    v.NamespaceName = namespace + ":" + version
    v.Description = description
    v.Severity = severity_map.get(alas.sev, "Unknown")
    v.Metadata = {
        "CVE": [],
    }

    if alas.cves:
        v.Metadata["CVE"] = [{"Name": cve} for cve in alas.cves]

    v.Link = alas.url
    for item in sorted(fixed_in):
        f = FixedIn()
        f.Name = item.pkg
        f.NamespaceName = v.NamespaceName
        f.VersionFormat = "rpm"
        f.Version = item.ver
        v.FixedIn.append(f)

    return v
